"""
SonicMesh - Acoustic Ultrasonic Data Transfer Library
-----------------------------------------------------

This package provides:
- High-level APIs for sending text and files over sound
- Ultrasonic FSK encoder (64-FSK)
- FFT-based ultrasonic decoder
- WAV utilities for saving/reading transmissions

Modules exposed at top-level:
    encode_message, encode_file
    transmit, transmitf
    decode_wav
    decode_signal, decode_file
"""

from .encoder import (
    encode_message,
    encode_file,
)

from .transmit import(
    transmit, transmitf,
)

from .decode_wav import decode_wav

from .decoder import (
    decode_signal, decode_file,
)

# re exporting key config consts
from .acoustic_config import(
    SAMPLE_RATE, 
    SYMBOL_DURATION,
    SYMBOL_BITS,
    FREQ_TABLE,
)

__all__ = [
    "encode_message", "encode_file",
    "transmit", "transmitf",
    "decode_wav",
    "decode_signal", "decode_file",
    "SAMPLE_RATE", "SYMBOL_DURATION",
    "SYMBOL_BITS", "FREQ_TABLE",
]


import sounddevice as sd
from .encoder import encode_message, encode_file
from scipy.io.wavfile import write

import numpy as np
from .acoustic_config import SAMPLE_RATE

def _save_wav(path, signal):
    """Convert float32 => 16 bit PCM & store wav"""
    scaled = np.int16(signal / np.max(np.abs(signal)) * 32767)
    write(path, SAMPLE_RATE, scaled)

#transmitting messages
def transmit(msg: str, save_file=True):

    #encoding the msg into ultrasonic tones
    signal = encode_message(msg)


    if save_file:
        # convert signal to 16bit pcm for wav
        _save_wav("message.wav", signal)

    #Play the signal
    sd.play(signal, SAMPLE_RATE)
    sd.wait()

#transmitting ACTUAL FILES
def transmitf(file_path: str, save_file=True):
    print("[SonicMesh]  Encoding file: ", file_path)

    #encode file into bitstr
    signal = encode_file(file_path)

    if save_file:
        _save_wav("file_message.wav", signal)

    sd.play(signal, SAMPLE_RATE)
    sd.wait()


import numpy as np
import lzma
import zlib
from .acoustic_config import(
    SAMPLE_RATE, SYMBOL_DURATION, AMPLITUDE, 
    FREQ_TABLE, CHUNK_SIZE, SYMBOL_BITS,
    SILENCE_BETWEEN_PACKETS, WINDOW_FUNCTION)


sync_indices = [63,0,63,0]
sync_bits = ''.join(f"{i:0{SYMBOL_BITS}b}" for i in sync_indices)
bits_per_symbol = SYMBOL_BITS

def encode_symbol(bitchunk: str):
    """
    encode 6bits (string) into an ultrasonic tone.
    """

    idx = int(bitchunk, 2)
    freq = FREQ_TABLE[idx]

    samples = int(SAMPLE_RATE*SYMBOL_DURATION)

    t = np.linspace(0, SYMBOL_DURATION, samples, endpoint=False)

    # windowed tone to cleaner FFT peak
    tone = AMPLITUDE*np.sin(2*np.pi*freq*t)
    if WINDOW_FUNCTION is not None:
        tone *= WINDOW_FUNCTION(len(t))

    return tone.astype(np.float32)
    
def encode_bits_fsk(bitstream: str):
    """Stream of bits => concatenatd ultrasonic tones + silence spacings"""
    tones = []
    silence = np.zeros(int(SAMPLE_RATE*SILENCE_BETWEEN_PACKETS), dtype=np.float32)
    

    for i in range(0, len(bitstream), bits_per_symbol):
        chunk = bitstream[i:i+bits_per_symbol]
        if len(chunk) < bits_per_symbol:
            chunk = chunk.ljust(bits_per_symbol,'0') #pad

        tone = encode_symbol(chunk)
        tones.append(tone) # it would hlep FFT distinguish symbols
        if SILENCE_BETWEEN_PACKETS > 0:
            tones.append(silence)

    signal =  np.concatenate(tones)
    return signal.astype(np.float32)


# ======================================
# FILE => PACKETS => BITSTREAM => AUDIO
# ======================================

def packetize_file(path):
    
    with open(path, "rb") as f:
        raw = f.read()

    compressed = lzma.compress(raw, preset=6)

    packets = []

    for i in range(0, len(compressed), CHUNK_SIZE):
        chunk = compressed[i:i + CHUNK_SIZE]

        packet = bytearray()
        packet += len(chunk).to_bytes(2, "big")
        packet += chunk
        packet += zlib.crc32(chunk).to_bytes(4, "big")

        packets.append(packet)

    return packets

# packets to binary string
def packets_to_bits(packets):
    bitstream = ""

    preamble_repeats = 8
    sync_bits = ''.join(f"{i:0{bits_per_symbol}b}" for i in sync_indices)
    for p in packets:
        bitstream += sync_bits * preamble_repeats
        for byte in p:
            bitstream += f"{byte:08b}"

    # pad bitstream to full symbols
    remainder = len(bitstream) % bits_per_symbol
    if remainder  != 0:
        pad_len= bits_per_symbol - remainder
        bitstream+= "0" * pad_len

    return bitstream


# high level file encoder
def encode_file(path):
    packets = packetize_file(path)
    bitstream = packets_to_bits(packets)
    return encode_bits_fsk(bitstream)

# ========================================
#    TEXT MESSAGE ENCODING (simple mode)
# ========================================

def encode_message(msg: str):
    data = msg.encode("utf-8")
    length = len(data)

    header = length.to_bytes(2,"big")

    payload = header + data

    bitstream = ''.join(f"{byte:08b}" for byte in payload)

    if len(bitstream) % bits_per_symbol != 0:
        pad_len = bits_per_symbol - (len(bitstream) % bits_per_symbol)
        bitstream += "0" * pad_len

    return encode_bits_fsk(bitstream)



import soundfile as sf
import numpy as np
from .decoder import decode_signal, decode_file

def bits_to_text(bitstream):
    """"Convert an 8-bit aligned bitstream into a UTF-8 string."""
    text = ""
    for i in range(0, len(bitstream),8):
        byte = bitstream[i:i+8]
        if len(byte) == 8:
            text+=chr(int(byte,2))
    return text
    

def decode_wav(wav_path: str, as_file: bool = False, output_file: str = None ):
    """
    Decode a SonicMesh-generated WAV file.
    
    Parameters
    ==========
    wav_path : str
        Path to the WAV file containing the encoded ultrasonic signal.
    as_file : bool, optional
        If True, treating the payload as a binary fine and reconstructing it.
        If False, decode the payload as UTF-8 text.
    output_file : str, optional
        Filepath to write the reconstructedd binary output, used only 
        when `as_file=True`.

    Returns
    =======
    str
        - when `as_file= False`: decoded text msg.
        - when `as_file= True `: path to the reconstructed file

    """

    data, sr = sf.read(wav_path)

    #flattening in case if its stereo
    signal = np.array(data).flatten()

    # Core decoder to convert audio -> bitstream.
    bits = decode_signal(signal)

    if as_file == True:
        if output_file== None:
            output_file = "received_file.bin"

        # Reconstructed file from the decoded bistream
        decode_file(bits, output_file)
        return output_file
    
    else:
        
        return bits_to_text(bits)
